/*
	File:		CDROMSample.c
	
	Description:	This sample demonstrates how to use IOKitLib to find CD-ROM media mounted on the 			system. It also shows how to open, read raw sectors from, and close the drive.
                
	Author:		gc

	Copyright: 	 Copyright 2000 Apple Computer, Inc. All rights reserved.
	
	Disclaimer:	IMPORTANT:  This Apple software is supplied to you by Apple Computer, Inc.
                        ("Apple") in consideration of your agreement to the following terms, and your
                        use, installation, modification or redistribution of this Apple software
                        constitutes acceptance of these terms.  If you do not agree with these terms,
                        please do not use, install, modify or redistribute this Apple software.

                        In consideration of your agreement to abide by the following terms, and subject
                        to these terms, Apple grants you a personal, non-exclusive license, under Apples
                        copyrights in this original Apple software (the "Apple Software"), to use,
                        reproduce, modify and redistribute the Apple Software, with or without
                        modifications, in source and/or binary forms; provided that if you redistribute
                        the Apple Software in its entirety and without modifications, you must retain
                        this notice and the following text and disclaimers in all such redistributions of
                        the Apple Software.  Neither the name, trademarks, service marks or logos of
                        Apple Computer, Inc. may be used to endorse or promote products derived from the
                        Apple Software without specific prior written permission from Apple.  Except as
                        expressly stated in this notice, no other rights or licenses, express or implied,
                        are granted by Apple herein, including but not limited to any patent rights that
                        may be infringed by your derivative works or by other works in which the Apple
                        Software may be incorporated.

                        The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO
                        WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
                        WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
                        PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND OPERATION ALONE OR IN
                        COMBINATION WITH YOUR PRODUCTS.

                        IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
                        CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
                        GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
                        ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION
                        OF THE APPLE SOFTWARE, HOWEVER CAUSED AND WHETHER UNDER THEORY OF CONTRACT, TORT
                        (INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN IF APPLE HAS BEEN
                        ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
				
	Change History (most recent first):
        
            <1>	 	08/16/00	gc		New sample.
        
*/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <dev/disk.h>
#include <errno.h>
#include <paths.h>
#include <sys/param.h>
#include <IOKit/IOKitLib.h>
#include <IOKit/IOBSD.h>
#include <IOKit/storage/IOMedia.h>
//#include <IOKit/storage/IOCDMedia.h>
// the above header is currently missing: this will be fixed in the future
#include <CoreFoundation/CoreFoundation.h>

#define kIOCDMediaClass	"IOCDMedia"	// this should be defined in a header somewhere but isn't currently
#define kIOMediaClass	"IOMedia"	// this should be defined in a header somewhere but isn't currently

enum {
    kCDSectorSizeWhole = 2352 	// this should be defined in a header somewhere but isn't currently
};

static kern_return_t FindEjectableCDMedia( io_iterator_t *mediaIterator );
static kern_return_t GetBSDPath( io_iterator_t mediaIterator, char *bsdPath, CFIndex maxPathSize );
static int OpenDrive( const char *bsdPath );
static Boolean ReadSector( int fileDescriptor );
static void CloseDrive( int fileDescriptor );

// Returns an iterator across all CD media (class IOCDMedia). Caller is responsible for releasing
// the iterator when iteration is complete.
kern_return_t FindEjectableCDMedia( io_iterator_t *mediaIterator )
{
    kern_return_t		kernResult; 
    mach_port_t			masterPort;
    CFMutableDictionaryRef	classesToMatch;

/*! @function IOMasterPort
    @abstract Returns the mach port used to initiate communication with IOKit.
    @discussion Functions that don't specify an existing object require the IOKit master port to be passed. This function obtains that port.
    @param size The task's bootstrap port, pass bootstrap_port from mach/mach_init.h.
    @param masterPort The master port is returned.
    @result A kern_return_t error code. */

    kernResult = IOMasterPort( bootstrap_port, &masterPort );
    if ( KERN_SUCCESS != kernResult )
        printf( "IOMasterPort returned %d\n", kernResult );
        
/*! @function IOServiceMatching
    @abstract Create a matching dictionary that specifies an IOService class match.
    @discussion A very common matching criteria for IOService is based on its class. IOServiceMatching will create a matching dictionary that specifies any IOService of a class, or its subclasses. The class is specified by C-string name.
    @param name The class name, as a const C-string. Class matching is successful on IOService's of this class or any subclass.
    @result The matching dictionary created, is returned on success, or zero on failure. The dictionary is commonly passed to IOServiceGetMatchingServices or IOServiceAddNotification which will consume a reference, otherwise it should be released with CFRelease by the caller. */

    // CD media are instances of class kIOCDMediaClass
    classesToMatch = IOServiceMatching( kIOCDMediaClass ); 
    if ( classesToMatch == NULL )
        printf( "IOServiceMatching returned a NULL dictionary.\n" );
    else {
/*!
	@function CFDictionarySetValue
	Sets the value of the key in the dictionary.
	@param theDict The dictionary to which the value is to be set. If this
		parameter is not a valid mutable CFDictionary, the behavior is
		undefined. If the dictionary is a fixed-capacity dictionary and
		it is full before this operation, and the key does not exist in
		the dictionary, the behavior is undefined.
	@param key The key of the value to set into the dictionary. If a key 
		which matches this key is already present in the dictionary, only
		the value is changed ("add if absent, replace if present"). If
		no key matches the given key, the key-value pair is added to the
		dictionary. If added, the key is retained by the dictionary,
		using the retain callback provided
		when the dictionary was created. If the key is not of the sort
		expected by the key retain callback, the behavior is undefined.
	@param value The value to add to or replace into the dictionary. The value
		is retained by the dictionary using the retain callback provided
		when the dictionary was created, and the previous value if any is
		released. If the value is not of the sort expected by the
		retain or release callbacks, the behavior is undefined.
*/

	CFDictionarySetValue( classesToMatch, CFSTR( kIOMediaEjectable ), kCFBooleanTrue ); 
        // Each IOMedia object has a property with key kIOMediaEjectable which is true if the media
        // is indeed ejectable. So add this property to the CFDictionary we're matching on. 
    }
    
    /*! @function IOServiceGetMatchingServices
        @abstract Look up registered IOService objects that match a matching dictionary.
        @discussion This is the preferred method of finding IOService objects currently registered by IOKit. IOServiceAddNotification can also supply this information and install a notification of new IOServices. The matching information used in the matching dictionary may vary depending on the class of service being looked up.
        @param masterPort The master port obtained from IOMasterPort().
        @param matching A CF dictionary containing matching information, of which one reference is consumed by this function. IOKitLib can contruct matching dictionaries for common criteria with helper functions such as IOServiceMatching, IOOpenFirmwarePathMatching.
        @param existing An iterator handle is returned on success, and should be released by the caller when the iteration is finished.
        @result A kern_return_t error code. */

    kernResult = IOServiceGetMatchingServices( masterPort, classesToMatch, mediaIterator );    
    if ( KERN_SUCCESS != kernResult )
        printf( "IOServiceGetMatchingServices returned %d\n", kernResult );
        
    return kernResult;
}
    
// Given an iterator across a set of CD media, return the BSD path to the
// next one. If none are found the path name is set to an empty string.
kern_return_t GetBSDPath( io_iterator_t mediaIterator, char *bsdPath, CFIndex maxPathSize )
{
    io_object_t		nextMedia;
    kern_return_t	kernResult = KERN_FAILURE;
    
    /*! @function IOIteratorNext
    @abstract Returns the next object in an iteration.
    @discussion This function returns the next object in an iteration, or zero if no more remain or the iterator is invalid.
    @param iterator An IOKit iterator handle.
    @result If the iterator handle is valid, the next element in the iteration is returned, otherwise zero is returned. */

    nextMedia = IOIteratorNext( mediaIterator );
    if ( nextMedia == NULL )
    {
        *bsdPath = '\0';
    }
    else {
        CFTypeRef	bsdPathAsCFString;
        
/*! @function IORegistryEntryCreateCFProperty
    @abstract Create a CF representation of a registry entry's property.
    @discussion This function creates an instantaneous snapshot of a registry entry property, creating a CF container analogue in the caller's task. Not every object available in the kernel is represented as a CF container; currently OSDictionary, OSArray, OSSet, OSSymbol, OSString, OSData, OSNumber, OSBoolean are created as their CF counterparts. 
    @param entry The registry entry handle whose property to copy.
    @param key A CFString specifying the property name.
    @param allocator The CF allocator to use when creating the CF container.
    @param options No options are currently defined.
    @result A CF container is created and returned the caller on success. The caller should release with CFRelease. */
        bsdPathAsCFString = IORegistryEntryCreateCFProperty( nextMedia, 
                                                             CFSTR( kIOBSDName ), 
                                                             kCFAllocatorDefault, 
                                                             0 );

        *bsdPath = '\0';
        if ( bsdPathAsCFString )
        {
            size_t devPathLength = strlen( _PATH_DEV );
            
            strcpy( bsdPath, _PATH_DEV );
            
            if ( CFStringGetCString( bsdPathAsCFString,
                                     bsdPath + devPathLength,
                                     maxPathSize - devPathLength, 
                                     kCFStringEncodingASCII ) )
            {
                printf( "BSD path: %s\n", bsdPath );
                kernResult = KERN_SUCCESS;
            }
            
            CFRelease( bsdPathAsCFString );
        }
    }
    
/*! @function IOObjectRelease
    @abstract Releases an object handle previously returned by IOKitLib.
    @discussion All objects returned by IOKitLib should be released with this function when access to them is no longer needed. Using the object after it has been released may or may not return an error, depending on how many references the task has to the same object in the kernel.
    @param object The IOKit object to release.
    @result A kern_return_t error code. */
    
    IOObjectRelease( nextMedia );
    
    return kernResult;
}

// Given the path to a CD drive, open the drive.
// Return the file descriptor associated with the device.
int OpenDrive( const char *bsdPath )
{
    int	fileDescriptor;
    
    fileDescriptor = open( bsdPath, O_RDONLY );
    
    if ( fileDescriptor == -1 )
        printf( "Error %d opening device %s.\n", errno, bsdPath );
    
    return fileDescriptor;
}

// Given the file descriptor for a whole-media CD device, read a sector drive.
// Return true if successful, otherwise false.
Boolean ReadSector( int fileDescriptor )
{
    char	buffer[ kCDSectorSizeWhole ];
    ssize_t	numBytes;
    
    numBytes = read( fileDescriptor, buffer, kCDSectorSizeWhole );
        
    return numBytes == kCDSectorSizeWhole ? true : false;
}

// Given the file descriptor for a device, close that device.
void CloseDrive( int fileDescriptor )
{
    close( fileDescriptor );
}

int main( int argc, char *argv[] )
{
    kern_return_t	kernResult; // on PowerPC this is an int (4 bytes)
/*
 *	error number layout as follows (see mach/error.h):
 *
 *	hi		 		       lo
 *	| system(6) | subsystem(12) | code(14) |
 */

    io_iterator_t	mediaIterator;
    char		bsdPath[ MAXPATHLEN ];
 
    kernResult = FindEjectableCDMedia( &mediaIterator );
    
    kernResult = GetBSDPath( mediaIterator, bsdPath, sizeof( bsdPath ) );
    
    // Now open the device we found, read a sector, and close the device
    if ( bsdPath[ 0 ] != '\0' )
    {
        int fileDescriptor;
        
        fileDescriptor = OpenDrive( bsdPath );
        
        if ( ReadSector( fileDescriptor ) )
            printf( "Sector read successfully.\n" );
        else
            printf( "Could not read sector.\n" );
            
        CloseDrive( fileDescriptor );
        printf( "Device closed.\n" );
    }
    else
        printf( "No ejectable CD media found.\n" );

    // Release the iterator.
    IOObjectRelease( mediaIterator );
        
    return 0;
}
